// Copyright (c) 2025 shibedrill
// SPDX-License-Identifier: GPL-3.0-or-later

#![no_std]
#![no_main]
#![feature(allocator_api)]

mod arch;
mod boot;
mod log;
mod memory;
mod panic;
mod params;
mod process;
mod resources;
mod syscall_runner;

use arch::current::*;
use arch::x86_64::serial::Serialport;
use limine::file::File;
use limine::memory_map::EntryType;
#[allow(unused_imports)]
use lzma_rs::lzma_decompress;
use spin::mutex::Mutex;
use syscall_runner::KERNEL_VERSION_MAJOR;
use syscall_runner::KERNEL_VERSION_MINOR;
use syscall_runner::KERNEL_VERSION_PATCH;

use crate::boot::*;
use crate::log::*;
use crate::memory::alloc::{boxed, collections::BTreeMap, format, string::*, vec};
use crate::params::*;

#[unsafe(no_mangle)]
unsafe extern "C" fn main() -> ! {
    // Assert supported bootloader version
    assert!(BASE_REVISION.is_supported());

    // TODO: This is hacky and will not work on any other machine.
    let qemu_serial = Serialport::new(0x3f8);

    // Local logger fn
    fn log(level: LogLevel, msg: &str) {
        LOGGER.lock().log(level, msg);
    }

    let mut params: BTreeMap<String, String> = BTreeMap::new();

    // Set up some stuff based on kernel params if we get any
    if let Some(executable_file_response) = FILE_REQUEST.get_response() {
        params = get_kernel_params(
            limine::file::File::string(executable_file_response.file()).to_bytes(),
        );

        // Set up logging level from params
        // Nothing we can do here if this fails since no log subscribers are initialized yet
        if let Some(level) = params.get("-loglevel")
            && let Ok(parsed_level) = LogLevel::try_from(level.as_str())
        {
            LOGGER.lock().level = parsed_level;
        }
        // Add subscribers to logger
        if let Some(device) = params.get("-logdev") {
            let log_device_list: vec::Vec<&str> = device.split(',').collect();
            if log_device_list.contains(&"display") {
                // Append display console to log subs
            }
            if log_device_list.contains(&"serial") {
                LOGGER
                    .lock()
                    .subscriber
                    .push(Mutex::new(boxed::Box::new(qemu_serial)));
            }
            log(LogLevel::Info, "Boot: Configured kernel logging devices.")
        }
        log(
            LogLevel::Info,
            &format!(
                "Boot: Kernel cmdline: {}",
                String::from_utf8_lossy(
                    limine::file::File::string(executable_file_response.file()).to_bytes()
                )
            ),
        );
        log(
            LogLevel::Info,
            &format!(
                "Boot: Kernel file path: {}",
                String::from_utf8_lossy(executable_file_response.file().path().to_bytes())
            ),
        );
    }

    log(
        LogLevel::Info,
        &format!(
            "Boot: Booting gila version {KERNEL_VERSION_MAJOR}.{KERNEL_VERSION_MINOR}.{KERNEL_VERSION_PATCH}"
        ),
    );

    log(LogLevel::Info, "Boot: Trans Rights!");

    if let Some(address_response) = ADDRESS_REQUEST.get_response() {
        log(
            LogLevel::Trace,
            &format!(
                "Boot: Kernel address: 0x{:0x}",
                address_response.physical_base()
            ),
        )
    }

    if let Some(framebuffer_response) = FRAMEBUFFER_REQUEST.get_response() {
        let fb = framebuffer_response.framebuffers().next().unwrap();
        log(LogLevel::Info, "Boot: Framebuffer response received");
        log(
            LogLevel::Trace,
            &format!(
                "Boot: Framebuffer dimensions: {}x{}, {}",
                fb.width(),
                fb.height(),
                fb.bpp()
            ),
        );
    }

    let _smp_response = MP_REQUEST.get_response();
    match _smp_response {
        None => log(LogLevel::Error, "MP: Multiprocessing response not received"),
        Some(resp) => {
            log(LogLevel::Info, "MP: Multiprocessing response received");
            log(
                LogLevel::Info,
                &format!("MP: {} CPUs found", resp.cpus().len()),
            );
        }
    }

    let _rsdp_response = RSDP_REQUEST.get_response();

    // TODO: Get the initramfs
    if let Some(module_response) = MODULE_REQUEST.get_response() {
        log(
            LogLevel::Info,
            &format!(
                "Boot: {} kernel modules found",
                module_response.modules().len()
            ),
        );
        if !module_response.modules().is_empty() {
            log(LogLevel::Info, "Boot: Kernel modules list:");
            for module in module_response.modules() {
                log(
                    LogLevel::Info,
                    &format!("\t{}", String::from_utf8_lossy(module.path().to_bytes())),
                )
            }
        }
        if let Some(irfs_path) = params.get("-initramfs") {
            log(
                LogLevel::Info,
                &format!("Boot: initramfs path requested: {irfs_path}"),
            );
            let initramfs_option: Option<&File> = {
                let mut result: Option<&File> = None;
                for file in module_response.modules() {
                    if &file.path().to_string_lossy().to_string() == irfs_path {
                        result = Some(file);
                    }
                }
                result
            };
            if let Some(_initramfs_file) = initramfs_option {
                log(LogLevel::Info, "Boot: Found initramfs at supplied path")
            }
        }
    }

    if let Some(mmap_response) = MEMMAP_REQUEST.get_response() {
        log(LogLevel::Info, "Boot: Memory map received");
        if !mmap_response.entries().is_empty() {
            log(LogLevel::Info, "Boot: Memory map entries list:");
            let mut usable: u64 = 0;
            let mut reclaimable: u64 = 0;
            for entry in mmap_response.entries() {
                log(
                    LogLevel::Info,
                    &format!(
                        "\t0x{:x} @ 0x{:x}: {}",
                        entry.length,
                        entry.base,
                        match entry.entry_type {
                            EntryType::ACPI_NVS => "ACPI (reserved)",
                            EntryType::ACPI_RECLAIMABLE => {
                                reclaimable += entry.length;
                                "ACPI (reclaimable)"
                            }
                            EntryType::BAD_MEMORY => "damaged/unusable",
                            EntryType::BOOTLOADER_RECLAIMABLE => {
                                reclaimable += entry.length;
                                "bootloader (reclaimable)"
                            }
                            #[allow(unreachable_patterns, deprecated)]
                            EntryType::EXECUTABLE_AND_MODULES | EntryType::KERNEL_AND_MODULES =>
                                "executable & modules",
                            EntryType::FRAMEBUFFER => "framebuffer",
                            EntryType::RESERVED => "reserved",
                            EntryType::USABLE => {
                                usable += entry.length;
                                "usable"
                            }
                            _ => "unidentified",
                        }
                    ),
                )
            }
            log(
                LogLevel::Info,
                &format!(
                    "Boot: Memory report: {usable} bytes usable, {reclaimable} bytes reclaimable, {} bytes ultimately available",
                    usable + reclaimable
                ),
            )
        }
    }

    loop {
        for _i in 0..100000000 {
            unsafe {
                core::arch::asm!("nop");
            }
        }
        log(LogLevel::Trace, "Heartbeat")
    }
}
